
/*
 * DREADERD/DNS.C - dns resolution and authentication task
 *
 *	DNS authenticator for new connections.
 *
 * (c)Copyright 1998, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution
 *    for specific rights granted.
 */

#include "defs.h"

#define MAXKPAIR 32

Prototype void DnsTask(int fd, const char *id);
Prototype void InstallAccessCache(void);
Prototype int ReadAccessCache(void);
Prototype void UpdateAuthDetails(ForkDesc *desc);
Prototype void SetAuthDetails(DnsRes *dres, char *which);
Prototype void PrintAuthDetails(DnsRes *dres);
Prototype void ClearOldAccessMap(void);

void DnsTest(DnsReq *dreq, DnsRes *dres, char *fqdn, char **ary, const char *ipname);
int fileAuthenticate(char *user, char *pass, char *conf);
void getAuthUser(const struct sockaddr_in *lsin, const struct sockaddr_in *rsin, char *ubuf, int ulen);
int readExactly(int fd, void *buf, int bytes);
void sigSegVDNS(int sigNo);
int readtoeof(int fd, void *buf, int bytes, int secs);
int readAccessLine(FILE *af, char *fname, AccessDef *adef);
#ifdef RADIUS_ENABLED
int radiusAuthenticate(char *user, char *pass, char *conf);
#endif
#ifdef CDB_ENABLED
extern int cdb_seek(int, char*, unsigned int, unsigned int*);
int cdbAuthenticate(char *user, char *pass, char *conf);
#endif

int DFd = -1;
MemPool	*DnsMemPool;

time_t Access_LastUpdate = 0;
time_t Cache_LastUpdate = 0;

AccessMap accessmap;
AccessMap oldaccessmap;

/* -------------------------------------------------------------------*/

/*
 * This is the process that loops handling DNS and authentication requests
 *
 * It listens in a UNIX SOCKET for a dreq structure and responds with a
 * dres structure.
 *
 */

void
DnsTask(int fd, const char *id)
{
    /*
     * read remote address to resolve.  Note: fd is not set to non-blocking
     */
    DnsReq dreq;
    int n;

    DFd = fd;

    if (CoreDebugOpt)
        signal(SIGSEGV, sigSegVDNS);

    /*
     * Make dummy call to lock buffer so we do not inefficiently free and
     * reallocate it for every DNS request.
     */
    (void)zalloc(&DnsMemPool, 8);

    
    /*
     * Loop on DNS requests, resolve them, and respond.
     */

    while ((n = readExactly(fd, &dreq, sizeof(dreq))) == sizeof(dreq)) {
	struct hostent *he;
	char *ipname;
	DnsRes dres;

	bzero(&dres, sizeof(dres));

	stprintf("reverse auth %s", inet_ntoa(dreq.dr_RSin.sin_addr));

	stprintf("dns lookup %s", inet_ntoa(dreq.dr_RSin.sin_addr));

	dres.dr_Addr = dreq.dr_RSin.sin_addr;
	dres.dr_Port = dreq.dr_RSin.sin_port;

	he = gethostbyaddr(
	    (char *)&dreq.dr_RSin.sin_addr,
	    sizeof(dreq.dr_RSin.sin_addr),
	    AF_INET
	);

	/*
	 * NOTE: we have to copy the fields returned by gethostbyaddr()
	 * because we will be doing other gethost*() calls in our tests and
	 * the original fields may get replaced out from under us.
	 */

	if (ReadAccessCache() == 1)
	    ClearOldAccessMap();

	ipname = zallocStr(&DnsMemPool, inet_ntoa(dreq.dr_RSin.sin_addr));
	snprintf(dres.dr_Host, sizeof(dres.dr_Host), "%s", ipname);

	if (he != NULL) {
	    char *hname;
	    char **haliases;
	    int nary;
	    int i;

	    for (nary = 0; he->h_aliases && he->h_aliases[nary]; ++nary)
		;

	    /*
	     * allocate
	     */
	    hname = zallocStr(&DnsMemPool, he->h_name);
	    haliases = zalloc(&DnsMemPool, sizeof(char *) * (nary + 1));
	    for (i = 0; i < nary; ++i) 
		haliases[i] = zallocStr(&DnsMemPool, he->h_aliases[i]);

	    /* Do forward DNS doublecheck */
	    if (TestForwardLookup(hname, &dreq.dr_RSin) < 0)
		dres.dr_DnsMismatch = 1;
	    else
		snprintf(dres.dr_Host, sizeof(dres.dr_Host), "%s", hname);

	    /*
	     * call test
	     */
	    if (DebugOpt)
		printf("call test, DnsTest %s\n", ipname);
	    DnsTest(&dreq, &dres, hname, haliases, ipname);

	    /*
	     * free
	     */
	    zfreeStr(&DnsMemPool, &hname);
	    for (i = 0; i < nary; ++i)
		zfreeStr(&DnsMemPool, &haliases[i]);
	    zfree(&DnsMemPool, haliases, sizeof(char *) * (nary + 1));
	} else {
	    if (DebugOpt)
		printf("ipname, DnsTest %s\n", ipname);
	    DnsTest(&dreq, &dres, NULL, NULL, ipname);
	}
	write(fd, &dres, sizeof(dres));
	zfreeStr(&DnsMemPool, &ipname);
	stprintf("dns idle");
    }
    logit(LOG_NOTICE, "DNS process exiting n=%d/%d", n, sizeof(dreq));
}

#ifdef CDB_ENABLED

/*
 * Search a CDB database for a matching username/passord
 *
 * Returns:	-1	failure
 *		 0	success
 *
 */
int cdbAuthenticate(char *user, char *pass, char *conf)
{
    unsigned int pwlen = 0;
    int r = -1, fd = open(conf, O_RDONLY);
    if (fd < 0) {
	logit(LOG_ERR, "unable to open auth cdb %s", conf);
	return(-1);
    }
    if (cdb_seek(fd, user, strlen(user), &pwlen) > 0) {
	if (pwlen > 0) {
	    char pw[64+1];
	    if (pwlen > sizeof(pw)-1) pwlen = sizeof(pw)-1;
	    if (read(fd, pw, pwlen) == pwlen) {
		pw[pwlen] = 0;
		if (!strcmp(pw, pass)) r = 0;
	    }
	}
    }
    close(fd);
    return r;
}

#endif /* CDB_ENABLED */

/*
 * Search a file of username:password pairs for a matching username/passord
 *
 * Returns:	-1	failure
 *		 0	success
 *
 */
int fileAuthenticate(char *user, char *pass, char *conf)
{
    FILE *fp;
    int r = -1;
    char buffer[1024];
    char *bptr;
    char *ptr;

    if (! ((fp = fopen(conf, "r")))) {
	logit(LOG_ERR, "unable to open auth file %s", conf);
	return(-1);
    }
    while (r && ! feof(fp)) {
	fgets(buffer, sizeof(buffer), fp);
	if ((ptr = strrchr(buffer, '\n'))) {
	    *ptr = '\0';
	}
	bptr = buffer;
	ptr = strsep(&bptr, ":");
	if (ptr && ! strcmp(ptr, user)) {
	    ptr = strsep(&bptr, ":");
	    if (ptr && ! strcmp(ptr, pass))
		r = 0;
	    break;
	}
    }
    fclose(fp);
    return(r);
}

#ifdef		RADIUS_ENABLED

/*
 * Send an authenticate request to a radius server
 *
 * Returns:	-1	failure
 *		 0	success
 */
int radiusAuthenticate(char *user, char *pass, char *conf)
{
    struct rad_handle *h;
    int rval;

    if (! ((h = rad_open()))) {
	logit(LOG_ERR, "unable to rad_auth_open");
	return(-1);
    }
    if (rad_config(h, conf) < 0) {
	logit(LOG_ERR, "rad_config(%s) failed: %s", conf, rad_strerror(h));
	rad_close(h);
	return(-1);
    }
    if (rad_create_request(h, RAD_ACCESS_REQUEST) < 0) {
	logit(LOG_ERR, "rad_create_request failed: %s", rad_strerror(h));
	rad_close(h);
	return(-1);
    }
    if (rad_put_string(h, RAD_USER_NAME, user) < 0) {
	logit(LOG_ERR, "rad_put_string(RAD_USER_NAME) failed: %s", rad_strerror(h));
	rad_close(h);
	return(-1);
    }
    if (rad_put_string(h, RAD_USER_PASSWORD, pass) < 0) {
	logit(LOG_ERR, "rad_put_string(RAD_USER_PASSWORD) failed: %s", rad_strerror(h));
	rad_close(h);
	return(-1);
    }
    if (((rval = rad_send_request(h))) < 0) {
	logit(LOG_ERR, "rad_send_request failed: %s", rad_strerror(h));
	rad_close(h);
	return(-1);
    }
    rad_close(h);

    if (rval == RAD_ACCESS_ACCEPT)
	rval = 0;
    else
	rval = -1;
    return(rval);
}
#endif

/*
 * DnsTest() - test access file entry against host.
 *
 *	fqdn	reverse lookup of name, if known, or NULL
 *	ary	aliases, if known, or NULL
 *	ipname	name of host as dotted quad
 */

void DnsTest(DnsReq *dreq, DnsRes *dres, char *fqdn, char **ary, const char *ipname)
{
    int AuthNeeded;
    int MatchedAuthNeeded = 0;
    int Authenticated = 0;
    int MatchedAuthenticated = 0;
    int MatchType;
    char buffer[1024];
    Vserver *vsp = NULL;
    AccessDef *alp;
    AccessDef *matchedalp = NULL;
    AccessDef accessdef;
    AccessDef matchedaccessdef;
    int acount;
    int DoneIdent = 0;
    int accessok;
    int accessfile;
    int firstmatch;
    FILE *af = NULL;

    /* This got trashed, so reset it */
    if (*dreq->dr_AuthUser) {
	dres->dr_RequireDns = 1;
	strcpy(dres->dr_AuthUser, dreq->dr_AuthUser);
	strcpy(dres->dr_AuthPass, dreq->dr_AuthPass);
    }

    for (vsp = accessmap.VServerList, acount = 0;
		acount < accessmap.VServerCount; acount++, vsp++) {
	if (dreq->dr_LSin.sin_addr.s_addr == vsp->vs_Interface) {
	    strcpy(dres->dr_VServer, vsp->vs_Name);
	    dres->dr_VServerDef = vsp;
            break;
	}
    }

    strncpy(buffer, PatLibExpand(DReaderHostsPat), sizeof(buffer) - 1);

    accessok = 1;
    accessfile = 0;

    /*
     * If we have specified a separate access file for a vserver,
     * then load it into memory for scanning.
     */
    if (*dres->dr_VServer && *dres->dr_VServerDef->vs_AccessFile) {
	accessfile = 1;
	strcpy(accessdef.ad_Reader, "DEFAULT");
	if (DebugOpt)
	    printf("Using accessfile: %s\n",
		PatLibExpand(dres->dr_VServerDef->vs_AccessFile));
	if ((af = fopen(PatLibExpand(dres->dr_VServerDef->vs_AccessFile), "r")) == NULL) {
	    syslog(LOG_CRIT, "Unable to open %s",	
		PatLibExpand(dres->dr_VServerDef->vs_AccessFile));
	    accessok = 0;
	}
    }

    /*
     * Not sure if this is still needed as we only ever call DnsTest()
     * from DnsTask(), which hopefully sets this up correctly.
     */
    if (!ipname)
	if (fqdn && IsIpAddr(fqdn))
	    ipname = fqdn;

    /*
     * For logging purposes, put something here
     */
    snprintf(dres->dr_Host, sizeof(dres->dr_Host), "%s", ipname);

    /*
     * Go through interations of the access list or read a vserver
     * access file and look for matches. Last match wins.
     *
     * We use the following logic for a match:
     *  - if no match -> fail
     *  - if last match with no auth required -> pass
     *  - if last match with auth required (and no auth details) and no
     *		previous match -> pass (requires auth to do things)
     *  - if last match with auth required (and no auth details) and prev
     *		match -> pass with prev match
     *  - if last match with auth required (with matching auth details) -> pass
     *  - if no auth required and no read,post,feed or status -> fail
     */

    alp = accessmap.AccessList;
    matchedalp = NULL;
    acount = 0;
    firstmatch = 1;
    while (accessok) {
	if (accessfile) {
	    if (readAccessLine(af,
		    (char *)PatLibExpand(dres->dr_VServerDef->vs_AccessFile),
		    &accessdef) != 1)
		break;
	    alp = &accessdef;
	} else {
	    if (!firstmatch)
		alp++;
	    firstmatch = 0;
	    if (acount++ >= accessmap.AccessCount || alp == NULL)
		break;
	}

	/*
	 * First we test the hostname or IP.
	 */

	MatchType = 0;
	if (IsIpAddr(alp->ad_Pattern)) {
	    if (!ipname || !*ipname)
		continue;
	    if (strchr(alp->ad_Pattern, '*') ||
				strchr(alp->ad_Pattern, '?')) {
		/* Wildmatch */
		if (WildCaseCmp(alp->ad_Pattern, ipname) != 0)
		    continue;
	    } else if (strchr(alp->ad_Pattern, '/')) {
		/* CIDR notation */
		if (!CidrMatch(alp->ad_Pattern, ipname))
		    continue;
	    } else {
		    /* Default - exact match */
		if (strcasecmp(ipname, alp->ad_Pattern) != 0)
		    continue;
	    }
	    snprintf(dres->dr_Host, sizeof(dres->dr_Host), "%s", ipname);
	    MatchType = 1;
	} else {
	    if (!fqdn || !*fqdn)
		continue;
	    if (strchr(alp->ad_Pattern, '*') ||
				strchr(alp->ad_Pattern, '?')) {
		/* Wildmatch */
		if (WildCaseCmp(alp->ad_Pattern, fqdn) != 0)
		    continue;
	    } else {
		/* Default - exact match */
		if (strcasecmp(fqdn, alp->ad_Pattern) != 0)
		    continue;
	    }
	    MatchType = 2;
	    snprintf(dres->dr_Host, sizeof(dres->dr_Host), "%s", fqdn);
	}
	if (*alp->ad_IdentUser) {
	    if (!DoneIdent && !*dres->dr_IdentUser) {
		/*
		 * this isn't an option where the user can actually
		 * enter something to pass the test; it is gained from
		 * identd info.  therefore it is an all-pass or all-fail
		 * type thing.
		 */
		getAuthUser(&dreq->dr_LSin, &dreq->dr_RSin, dres->dr_IdentUser,
			sizeof(dres->dr_IdentUser));
		DoneIdent = 1;
	    }
	    /* We don't need to go further if we fail the ident check */
	    if (strcmp(alp->ad_IdentUser, dres->dr_IdentUser) != 0)
		continue;
	}

	/*
	 *  Now me are matching an access line, check some of the
	 *  user details
	 */

	if (DebugOpt)
	    printf("Matched access line: %s%s%s (%s)\n",
			*alp->ad_IdentUser ? alp->ad_IdentUser : "",
			*alp->ad_IdentUser ? "@" : "",
			alp->ad_Pattern,
			alp->ad_Reader);

	SetAuthDetails(dres, alp->ad_Reader);

	/*
	 * Do we log with the hostname
	 */
	if (dres->dr_ReaderDef->rd_UseVerifiedDns && fqdn && *fqdn)
	    snprintf(dres->dr_Host, sizeof(dres->dr_Host), "%s", fqdn);

	/*
	 *  We can optionally deny an entry that doesn't have a DNS entry
	 */
	if (dres->dr_ReaderDef->rd_DenyNoDns && (!fqdn || !*fqdn)) {
	    logit(LOG_ERR, "Denying host with no DNS entry: %s", ipname);
	    continue;
	}

	/*
	 * If we mismatch Fwd/Rev and we matched against a domain
	 * then reject the connection for security reasons
	 */
	if (dres->dr_DnsMismatch && (dres->dr_ReaderDef->rd_DenyMismatchedDns
			 || (dres->dr_DnsMismatch && MatchType != 1))) {
	    if (DebugOpt)
		printf("Not allowing DNS Fwd/Rev Mismatch for: %s matching %s%s%s\n",
			ipname,
			*alp->ad_IdentUser ? alp->ad_IdentUser : "",
			*alp->ad_IdentUser ? "@" : "",
			alp->ad_Pattern);
	    continue;
	}

	/*
	 * We can optionally use the FQDN for logging even when
	 * matched with an IP
	 */
	if (dres->dr_ReaderDef->rd_UseVerifiedDns && !dres->dr_DnsMismatch &&
						fqdn && *fqdn)
	    snprintf(dres->dr_Host, sizeof(dres->dr_Host), "%s", fqdn);
		

	/*
	 * Ok.  Now at this point, we either don't have a match on this
	 * line, in which case we've already skipped to the next line,
	 * or we have a match...  but we may still need to authenticate.
	 */

	Authenticated = 0;
	AuthNeeded = 0;

	if (*dres->dr_AuthDef->au_Radius ||
		*dres->dr_AuthDef->au_File ||
		*dres->dr_AuthDef->au_Cdb ||
		*dres->dr_AuthDef->au_External ||
		*dres->dr_AuthDef->au_User)
	    AuthNeeded = 1;

	if (dres->dr_AuthDef->au_Ident && !DoneIdent && !*dres->dr_IdentUser) {
	    /*
	     * this isn't an option where the user can actually
	     * enter something to pass the test; it is gained from
	     * identd info.  therefore it is an all-pass or all-fail
	     * type thing.
	     */
	    getAuthUser(&dreq->dr_LSin, &dreq->dr_RSin, dres->dr_IdentUser,
			sizeof(dres->dr_IdentUser));
	    DoneIdent = 1;
	}

	/*
	 * The user has used the AUTHINFO comamnd - check the details
	 * against various authentication mechanisms.
	 *
	 * All methods specified in the auth block are used and any of
	 * them can match
	 */

	if (AuthNeeded) {

	     char user[128];
	     int authok;

	     strncpy(user, dreq->dr_AuthUser, sizeof(user) - 1 );
	     authok = 1;

	    /*
	     * The realm is anything after the '@' in a username
	     * Make sure this matches, otherwise the others won't
	     * match.
	     */

	    if (*dres->dr_AuthDef->au_Realm) {
		char *p;
		authok = 0;
		p = strchr(user, '@');
		if (p) {
		    *p++ = '\0';
		    if (strcmp(p, dres->dr_AuthDef->au_Realm) == 0)
			authok = 1;
		}
	    }

	    /*
	     * If the user was set in auth and we have Ident, we just compare
	     */

	    if (authok && DoneIdent && *dres->dr_IdentUser &&
			*dres->dr_AuthDef->au_User &&
			!strcmp(dres->dr_IdentUser, dres->dr_AuthDef->au_User))  {
		Authenticated = 2;
		if (DebugOpt)
		    printf("Authenticated via ident+user\n");
	    }

#ifdef	RADIUS_ENABLED

	    /*
	     * Check user+pass with radius server
	     */

	    if (authok && *dres->dr_AuthDef->au_Radius && *user &&
				*dreq->dr_AuthPass) {
		if (radiusAuthenticate(user, dreq->dr_AuthPass,
			dres->dr_AuthDef->au_Radius) == 0) {
		    Authenticated = 1;
		    if (DebugOpt)
			printf("Authenticated via radius\n");
		} else {
		    Authenticated = -1;
		    if (DebugOpt)
			printf("Authentication via radius FAILED\n");
		}
	    }
#else
	    if (authok && *dres->dr_AuthDef->au_Radius && *user &&
			*dreq->dr_AuthPass) {
		Authenticated = -1;
		if (DebugOpt)
		    printf("Authentication via radius FAILED\n");
		syslog(LOG_ERR, "Radius authentication requested, but not enabled");
	    }
#endif

	    /*
	     * Check a user+pass specified in the auth config
	     */

	    if (authok && *dres->dr_AuthDef->au_User && *user &&
			*dres->dr_AuthDef->au_Pass && *dreq->dr_AuthPass) {
		if (!strcmp(user, dres->dr_AuthDef->au_User) &&
			!strcmp(dreq->dr_AuthPass, dres->dr_AuthDef->au_Pass)) {
		    Authenticated = 1;
		    if (DebugOpt)
			printf("Authenticated via user+pass\n");
		} else {
		    Authenticated = -1;
		    if (DebugOpt)
			printf("Authentication via user+pass FAILED\n");
		}
	    }

	    /*
	     * Check for the user+pass in a file
	     */

	    if (authok && *dres->dr_AuthDef->au_File && *user &&
			*dreq->dr_AuthPass) {
		if (fileAuthenticate(user, dreq->dr_AuthPass,
			dres->dr_AuthDef->au_File) == 0) {
		    Authenticated = 1;
		    if (DebugOpt)
			printf("Authenticated via file\n");
		} else {
		    Authenticated = -1;
		    if (DebugOpt)
			printf("Authentication via file FAILED\n");
		}
	    }

#ifdef CDB_ENABLED

	    /*
	     * Check for the user+pass in a CDB
	     */

	    if (authok && *dres->dr_AuthDef->au_Cdb && *user &&
			*dreq->dr_AuthPass) {
		if (cdbAuthenticate(user, dreq->dr_AuthPass,
			dres->dr_AuthDef->au_Cdb) == 0) {
		    Authenticated = 1;
		    if (DebugOpt)
			printf("Authenticated via cdb\n");
		} else {
		    Authenticated = -1;
		    if (DebugOpt)
			printf("Authentication via cdb FAILED\n");
		}
	    }

#else
	    if (authok && *dres->dr_AuthDef->au_Cdb && *user &&
			*dreq->dr_AuthPass) {
		Authenticated = -1;
		if (DebugOpt)
		    printf("Authentication via cdb FAILED\n");
		syslog(LOG_ERR, "CDB authentication requested, but not enabled");
	    }
#endif /* CDB_ENABLED */

	}
	if (!AuthNeeded ||
		/* AUTHINFO and matched */
		(Authenticated > 0 && *dreq->dr_AuthPass) ||
		/* Ident + user only */
		(Authenticated == 2) ||
		/* AUTHINFO and not failed auth, but no other match */
		(Authenticated == 0 && (matchedalp == NULL))
	) {
	if (accessfile) {
	    memcpy(&matchedaccessdef, &accessdef, sizeof(accessdef));
	    matchedalp = &matchedaccessdef;
	    MatchedAuthenticated = Authenticated;
	    MatchedAuthNeeded = AuthNeeded;
	} else {
	    matchedalp = alp;
	    MatchedAuthenticated = Authenticated;
	    MatchedAuthNeeded = AuthNeeded;
	}
	}
    }
    if (accessfile)
	fclose(af);
    if (matchedalp == NULL) {
	if (DebugOpt)
	    printf("No access match\n");
	dres->dr_Code = 0;
	return;
    }
    if (*dreq->dr_AuthPass && MatchedAuthenticated <= 0) {
	if (DebugOpt)
	    printf("Authentication failed\n");
	dres->dr_Code = 0;
	return;
    }

    strcpy(dres->dr_ReaderName, matchedalp->ad_Reader);

    SetAuthDetails(dres, matchedalp->ad_Reader);

    /*
     * Now we have checked against all the access lines and matchedalp
     * points to the last match
     */

    if (DebugOpt)
	PrintAuthDetails(dres);

    /*
     * okay, if this was an authentication request, and we didn't auth,
     * then return a failure
     */
    dres->dr_Flags &= ~DF_AUTHREQUIRED;
    if (MatchedAuthNeeded && !MatchedAuthenticated) {
	dres->dr_Flags |= DF_AUTHREQUIRED;
    }

    dres->dr_Code = 1;
    if ((dres->dr_Flags & (DF_FEED|DF_READ|DF_POST|DF_STATUS)) == 0)
	dres->dr_Code = 0;
}

/*
 * IDENT
 *
 * getAuthUser() - authenticate the remote username by connecting to port
 *		   113 and requesting the user id of the remote port.  This
 *		   is used by X-Trace: and some authentication.
 */

void
getAuthUser(const struct sockaddr_in *plsin, const struct sockaddr_in *prsin, char *ubuf, int ulen)
{
    int cfd;
    int lport;
    int rport;
    int n;
    char buf[256];
    struct sockaddr_in lsin = *plsin;
    struct sockaddr_in rsin = *prsin;

    rport = ntohs(rsin.sin_port);
    lport = ntohs(lsin.sin_port);

    lsin.sin_port = 0;
    lsin.sin_family = AF_INET;

    if ((cfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        return;
    }
    if (bind(cfd, (struct sockaddr *)&lsin, sizeof(lsin)) < 0) {
	perror("bind");
        close(cfd);
        return;
    }
    rsin.sin_port = htons(113);
    rsin.sin_family = AF_INET;

    /*
     * Do asynchronous connection with 10 second timeout.  The timeout is
     * necessary to deal with broken clients or firewalls.
     */

    fcntl(cfd, F_SETFL, O_NONBLOCK);
    errno = 0;
    if (connect(cfd, (struct sockaddr *)&rsin, sizeof(rsin)) < 0) {
	fd_set wfds;
	struct timeval tv = { 10, 0 };

	if (errno != EINPROGRESS) {
	    close(cfd);
	    return;
	}
	FD_ZERO(&wfds);
	FD_SET(cfd, &wfds);
	if (select(cfd + 1, NULL, &wfds, NULL, &tv) == 0) {
	    close(cfd);
	    cfd = -1;
	}
    }

    /*
     * Send query, interpret response
     *
     * write() may fail if connect() failed to establish a connection.
     */

    if (cfd >= 0) {
	fcntl(cfd, F_SETFL, 0);
	sprintf(buf, "%d, %d\r\n", rport, lport);
	if (DebugOpt)
	    printf("IDENTD COMMAND %s\n", buf);
	if (write(cfd, buf, strlen(buf)) != strlen(buf)) {
	    n = -1;
	} else {
	    fcntl(cfd, F_SETFL, O_NONBLOCK);
	    n = readtoeof(cfd, buf, sizeof(buf) - 1, 10);
	}
    } else {
	n = -1;
    }
    if (n > 0) {
	char *uid;

        buf[n] = 0;

	if (DebugOpt)
	    printf("IDENTD RESPONSE %s\n", buf);

        if ((uid = strstr(buf, "USERID")) != NULL) {
            uid = strchr(uid + 1, ':');
            if (uid) {
                uid = strchr(uid + 1, ':');
                if (uid) {
                    uid = strtok(uid + 1, "\r\n: \t");
                    if (uid && strlen(uid) < ulen) {
                        strcpy(ubuf, uid);
			SanitizeString(ubuf);
                    }
                }
            }
        }
    }
    if (cfd >= 0)
	close(cfd);
}

int
readExactly(int fd, void *buf, int bytes)
{
    int r = 0;

    while (bytes > 0) {
	int n = read(fd, buf, bytes);

	if (n <= 0) {
	    if (n < 0 && r == 0)
		r = -1;
	    break;
	}
	buf = (char *)buf + n;
	bytes -= n;
	r += n;
    }
    return(r);
}

int
readtoeof(int fd, void *buf, int bytes, int secs)
{
    fd_set rfds;
    int r = 0;

    FD_ZERO(&rfds);

    for (;;) {
	struct timeval tv = { 0, 0 };
	int n;

	tv.tv_sec = secs;
	FD_SET(fd, &rfds);

	if (select(fd + 1, &rfds, NULL, NULL, &tv) == 0)
	    break;
	errno = 0;
	n = read(fd, buf, bytes);
	if (n == 0)
	    break;
	if (n < 0) {
	    if (errno == EWOULDBLOCK ||
		errno == EINTR ||
		errno == EAGAIN
	    ) {
		continue;
	    }
	    break;
	}
	r += n;
	buf = (char *)buf + n;
	bytes -= n;
    }
    return(r);
}

void
sigSegVDNS(int sigNo)
{
    if (DFd >= 0)
        close(DFd);
    nice(20);
    for (;;)
        ;
}   

int
readAccessLine(FILE *af, char *fname, AccessDef *adef)
{
    static int needread = 1;
    static int linecount = 0;
    static char buf[8192] = "";
    static char *opt;
    static char *val;
    char *p;

    while (needread) {
    	if (fgets(buf, sizeof(buf), af) == NULL)
	    return(0);
	linecount++;
	opt = buf;
	while (isspace((int)*opt))
	    opt++;
	if (*opt == '#' || *opt == '\n' || *opt == '\0')
	    continue;
	p = opt + 6;
	*p = '\0';
	if (strcmp(opt, "access") == 0) {
	   opt += 7;
	   while (isspace((int)*opt))
		opt++;
	   bzero(adef, sizeof(AccessDef));
	   break;
	} else {
	    syslog(LOG_ERR, "%s: Invalid command %s in line %d",
				fname, opt, linecount);
	    return(0);
	}
    }
    /*
     * At this point we should be pointing to the wildmat pattern
     */
    if (opt == NULL) {
	syslog(LOG_ERR, "%s: Missing wildmat in line %d",
				fname, linecount);
	needread = 1;
	return(0);
    }
    if (needread) {
	val = strtok(opt, " \t");
	val = strtok(NULL, " \t");
	if (val == NULL) {
	    syslog(LOG_ERR, "%s: Invalid readerdef in line %d",
					fname, linecount);
	    needread = 1;
	    return(0);
	}
	while (isspace((int)*val))
	    val++;
	if (*val == '#' || *val == '\n' || *val == '\0') {
	    syslog(LOG_ERR, "%s: Invalid readerdef in line %d",
					fname, linecount);
	    needread = 1;
	    return(0);
	}
	p = strchr(val, '\n');
	if (p)
	    *p = '\0';
	strcpy(adef->ad_Reader, val);
    }
    p = strchr(opt, ',');
    if (p != NULL) {
	*p++ = '\0';
	needread = 1;
    } else {
	p = opt;
    }
    {
	/*
	 *  Check for identuser@
	 */
	char *q;
	if (*opt && (q = strchr(opt, '@')) != NULL) {
	    *q++ = '\0';
	    strncpy(adef->ad_IdentUser, opt,
			sizeof(adef->ad_IdentUser) - 1);
	    opt = q;
	}
    }
    if (*opt && strlen(opt) < sizeof(adef->ad_Pattern)) {
	strcpy(adef->ad_Pattern, opt);
	opt = p;
	return(1);
    }
    syslog(LOG_ERR, "%s: Bad length for access %s in line %d",
					fname, opt, linecount);
    needread = 1;
    return(0);
}

/*
 * Install a new copy of the access cache
 * This is done by creating the new cache files first, locking and then
 * renaming the files.
 */

void
InstallAccessCache() {
    struct stat st;
    char buf[1024];
    int linecount = 0;
    int tfp;
    FILE *fp;
    FILE *vsf;
    FILE *grf;
    FILE *auf;
    FILE *rdf;
    FILE *raf;
    char *opt;
    char *val;
    Vserver vserver;
    GroupDef groupdef;
    AuthDef authdef;
    ReaderDef readerdef;
    GroupList *groupptr = NULL;
    AccessDef accessdef;
    int copt = OPT_NONE;
    char vsnewbuf[PATH_MAX];
    char grnewbuf[PATH_MAX];
    char aunewbuf[PATH_MAX];
    char rdnewbuf[PATH_MAX];
    char ranewbuf[PATH_MAX];

    if ((stat(PatLibExpand(DReaderAccessPat), &st) == 0) &&
		(st.st_mtime <= Access_LastUpdate))
	return;

    syslog(LOG_INFO, "Installing access cache");

    if ((fp = fopen(PatLibExpand(DReaderAccessPat), "r")) == NULL) {
	syslog(LOG_CRIT, "Unable to open %s", PatLibExpand(DReaderAccessPat));
	return;
    }
    if (fstat(fileno(fp), &st) != 0) {
	fclose(fp);
	return;
    }
    Access_LastUpdate = st.st_mtime;

    snprintf(vsnewbuf, sizeof(vsnewbuf), "%s.new",
				PatDbExpand(DRVserverCachePat));
    snprintf(grnewbuf, sizeof(grnewbuf), "%s.new",
				PatDbExpand(DRGroupCachePat));
    snprintf(aunewbuf, sizeof(aunewbuf), "%s.new",
				PatDbExpand(DRAuthCachePat));
    snprintf(rdnewbuf, sizeof(rdnewbuf), "%s.new",
				PatDbExpand(DRReaderCachePat));
    snprintf(ranewbuf, sizeof(ranewbuf), "%s.new",
				PatDbExpand(DRAccessCachePat));

    if ((vsf = fopen(vsnewbuf, "w+")) == NULL) {
	syslog(LOG_CRIT, "Unable to create %s", vsnewbuf);
	fclose(fp);
	return;
    }
    if ((grf = fopen(grnewbuf, "w+")) == NULL) {
	syslog(LOG_CRIT, "Unable to create %s", grnewbuf);
	fclose(fp);
	fclose(vsf);
	return;
    }
    if ((auf = fopen(aunewbuf , "w+")) == NULL) {
	syslog(LOG_CRIT, "Unable to create %s", aunewbuf);
	fclose(fp);
	fclose(vsf);
	fclose(grf);
	return;
    }
    if ((rdf = fopen(rdnewbuf, "w+")) == NULL) {
	syslog(LOG_CRIT, "Unable to create %s", rdnewbuf);
	fclose(fp);
	fclose(vsf);
	fclose(grf);
	fclose(auf);
	return;
    }

    if ((raf = fopen(ranewbuf, "w+")) == NULL) {
	syslog(LOG_CRIT, "Unable to create %s", ranewbuf);
	fclose(fp);
	fclose(vsf);
	fclose(grf);
	fclose(auf);
	fclose(rdf);
	return;
    }

    /* The first entry is the default */

    bzero(&vserver, sizeof(vserver));
    strcpy(vserver.vs_Name, "DEFAULT");
    strcpy(vserver.vs_HostName, ReportedHostName);
    strcpy(vserver.vs_ClusterName, ReportedHostName);
    strcpy(vserver.vs_PostPath, PathHost);
    strcpy(vserver.vs_NewsAdm, NewsAdmin);
    fwrite(&vserver, sizeof(vserver), 1, vsf);

    fwrite("DEFAULT\0", 8, 1, grf);
    groupdef.gr_Count = 1;
    fwrite("1\0", 2, 1, grf);
    fwrite("*\0", 2, 1, grf);

    bzero(&authdef, sizeof(authdef));
    strcpy(authdef.au_Name, "DEFAULT");
    fwrite(&authdef, sizeof(authdef), 1, auf);

    bzero(&readerdef, sizeof(readerdef));
    strcpy(readerdef.rd_Name, "DEFAULT");
    strcpy(readerdef.rd_Auth, "DEFAULT");
    strcpy(readerdef.rd_Groups, "DEFAULT");
    strcpy(readerdef.rd_Vserver, "DEFAULT");
    fwrite(&readerdef, sizeof(readerdef), 1, rdf);

    while (fgets(buf, sizeof(buf), fp) != NULL) {
	linecount++;
	opt = buf;
	while (isspace((int)*opt))
	    opt++;
	opt = strtok(opt, " \t\n");
	if (opt == NULL || *opt == '#' || *opt == '\n' || *opt == '\0')
	    continue;
	val = strtok(NULL, "\n");
	if (val == NULL || *val == '#')
	    val = "";
	else
	    while (isspace((int)*val))
		val++;

	if (strcmp(opt, "end") == 0) {
	    switch (copt) {
		case OPT_NONE:
		    syslog(LOG_ERR, "Found 'end' without define in line %d", linecount);
		    break;
		case OPT_VSERVER:
		    if (!*vserver.vs_ClusterName)
			strcpy(vserver.vs_ClusterName, vserver.vs_HostName);
		    if (strcmp(vserver.vs_Name, "DEFAULT") == 0) {
			long fpos;
			fpos = ftell(vsf);
			fseek(vsf, 0L, SEEK_SET);
			fwrite(&vserver, sizeof(vserver), 1, vsf);
			fseek(vsf, fpos, SEEK_SET);
		    } else {
			fwrite(&vserver, sizeof(vserver), 1, vsf);
		    }
		    bzero(&vserver, sizeof(vserver));
		    break;
		case OPT_GROUPS:
		    fwrite(groupdef.gr_Name, strlen(groupdef.gr_Name) + 1, 1, grf);
		    zfreeStr(&DnsMemPool, &groupdef.gr_Name);
		    {
			char buf[10];
			struct GroupList *gpp;
			struct GroupList *gp;

			sprintf(buf, "%d", groupdef.gr_Count);
			fwrite(&buf, strlen(buf) + 1, 1, grf);
			gp = groupdef.gr_Groups;
			while (gp != NULL) {
			    fwrite(gp->group, strlen(gp->group) + 1, 1, grf);
			    gpp = gp;
			    gp = gp->next;
			    zfreeStr(&DnsMemPool, &gpp->group);
			    zfree(&DnsMemPool, gpp, sizeof(GroupList));
			}
		    }
		    bzero(&groupdef, sizeof(groupdef));
		    break;
		case OPT_AUTH:
		    if (strcmp(authdef.au_Name, "DEFAULT") == 0) {
			long fpos;
			fpos = ftell(auf);
			fseek(auf, 0L, SEEK_SET);
			fwrite(&authdef, sizeof(authdef), 1, auf);
			fseek(auf, fpos, SEEK_SET);
		    } else {
			fwrite(&authdef, sizeof(authdef), 1, auf);
		    }
		    bzero(&authdef, sizeof(authdef));
		    break;
		case OPT_READERGRP:
		    if (strcmp(readerdef.rd_Name, "DEFAULT") == 0) {
			long fpos;
			fpos = ftell(rdf);
			fseek(rdf, 0L, SEEK_SET);
			fwrite(&readerdef, sizeof(readerdef), 1, rdf);
			fseek(rdf, fpos, SEEK_SET);
		    } else {
			fwrite(&readerdef, sizeof(readerdef), 1, rdf);
		    }
		    bzero(&readerdef, sizeof(readerdef));
		    break;
	    }
	    copt = OPT_NONE;
	    continue;
	}
	if (!*val) {
	    syslog(LOG_ERR, "dreader.access: Missing value/label for %s in line %d", opt, linecount);
	    Access_LastUpdate = 0;
	    continue;
	}
	/* Handle the definition options */
	switch (copt) {
	    case OPT_NONE: break;
	    case OPT_VSERVER:
		if (*val && strcmp(opt, "clustername") == 0) {
		    strncpy(vserver.vs_ClusterName, val, sizeof(vserver.vs_ClusterName) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "hostname") == 0) {
		    strncpy(vserver.vs_HostName, val, sizeof(vserver.vs_HostName) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "postpath") == 0) {
		    strncpy(vserver.vs_PostPath, val, sizeof(vserver.vs_PostPath) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "newsadmin") == 0) {
		    strncpy(vserver.vs_NewsAdm, val, sizeof(vserver.vs_NewsAdm) - 1);
		    continue;
		}
		if (*val && ((strcmp(opt, "organisation") == 0) ||
			(strcmp(opt, "organization") == 0))) {
		    strncpy(vserver.vs_Org, val, sizeof(vserver.vs_Org) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "abuseto") == 0) {
		    strncpy(vserver.vs_AbuseTo, val, sizeof(vserver.vs_AbuseTo) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "cryptpw") == 0) {
		    strncpy(vserver.vs_CryptPw, val, sizeof(vserver.vs_CryptPw) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "accessfile") == 0) {
		    strncpy(vserver.vs_AccessFile, val, sizeof(vserver.vs_AccessFile) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "interface") == 0) {
		    vserver.vs_Interface = inet_addr(val);
		    continue;
		}
		if (*val && strcmp(opt, "noxrefhostupdate") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			vserver.vs_NoXrefHostUpdate = 1;
		    else
			vserver.vs_NoXrefHostUpdate = 0;
		    continue;
		}
		syslog(LOG_ERR, "dreader.access: Invalid vserverdef option : %s in line %d\n", opt, linecount);
		continue;
	    case OPT_GROUPS:
		if (*val && strcmp(opt, "group") == 0) {
		    if (groupdef.gr_Groups == NULL) {
			groupptr = zalloc(&DnsMemPool, sizeof(GroupList));
			groupptr->group = zallocStr(&DnsMemPool, val);
			groupptr->next = NULL;
			groupdef.gr_Groups = groupptr;
			groupdef.gr_Count++;
		    } else {
			groupptr->next = zalloc(&DnsMemPool, sizeof(GroupList));
			groupptr = groupptr->next;
			groupptr->group = zallocStr(&DnsMemPool, val);
			groupptr->next = NULL;
			groupdef.gr_Count++;
		    }
		    continue;
		}
		syslog(LOG_ERR, "dreader.access: Invalid groupdef option : %s in line %d\n", opt, linecount);
		continue;
	    case OPT_AUTH:
		if (*val && strcmp(opt, "file") == 0) {
		    strncpy(authdef.au_File, val, sizeof(authdef.au_File) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "cdb") == 0) {
		    strncpy(authdef.au_Cdb, val, sizeof(authdef.au_Cdb) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "external") == 0) {
		    strncpy(authdef.au_External, val, sizeof(authdef.au_External) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "radius") == 0) {
		    strncpy(authdef.au_Radius, val, sizeof(authdef.au_Radius) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "user") == 0) {
		    strncpy(authdef.au_User, val, sizeof(authdef.au_User) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "pass") == 0) {
		    strncpy(authdef.au_Pass, val, sizeof(authdef.au_Pass) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "realm") == 0) {
		    strncpy(authdef.au_Realm, val, sizeof(authdef.au_Realm) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "ident") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			authdef.au_Ident = 1;
		    else
			authdef.au_Ident = 0;
		    continue;
		}
		syslog(LOG_ERR, "dreader.access: Invalid authdef option : %s in line %d\n", opt, linecount);
		continue;
	    case OPT_READERGRP:
		if (*val && strcmp(opt, "auth") == 0) {
		    strncpy(readerdef.rd_Auth, val, sizeof(readerdef.rd_Auth) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "groups") == 0) {
		    strncpy(readerdef.rd_Groups, val, sizeof(readerdef.rd_Groups) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "vserver") == 0) {
		    strncpy(readerdef.rd_Vserver, val, sizeof(readerdef.rd_Vserver) - 1);
		    continue;
		}
		if (*val && strcmp(opt, "read") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_Read = 1;
		    else
			readerdef.rd_Read = 0;
		    continue;
		}
		if (*val && strcmp(opt, "post") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_Post = 1;
		    else
			readerdef.rd_Post = 0;
		    continue;
		}
		if (*val && strcmp(opt, "feed") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_Feed = 1;
		    else
			readerdef.rd_Feed = 0;
		    continue;
		}
		if (*val && strcmp(opt, "status") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_Status = 1;
		    else
			readerdef.rd_Status = 0;
		    continue;
		}
		if (*val && strcmp(opt, "controlpost") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_ControlPost = 1;
		    else
			readerdef.rd_ControlPost = 0;
		    continue;
		}
		if (*val && strcmp(opt, "maxconn") == 0) {
		    readerdef.rd_MaxConnTotal = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "maxconnperhost") == 0) {
		    readerdef.rd_MaxConnPerHost = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "maxconnperuser") == 0) {
		    readerdef.rd_MaxConnPerUser = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "maxconnpergroup") == 0) {
		    readerdef.rd_MaxConnPerGroup = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "maxconnpervs") == 0) {
		    readerdef.rd_MaxConnPerVs = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "ratelimit") == 0) {
		    readerdef.rd_RateLimit = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "ratelimittax") == 0) {
		    readerdef.rd_RateLimitTax = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "pathcomponents") == 0) {
		    readerdef.rd_PathComponents = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "idletimeout") == 0) {
		    readerdef.rd_IdleTimeout = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "sessiontimeout") == 0) {
		    readerdef.rd_SessionTimeout = atoi(val);
		    continue;
		}
		if (*val && strcmp(opt, "allowdnsmismatch") == 0) {
		    /* Historical option - ignore it */
		    syslog(LOG_NOTICE, "dreader.access: Ignoring historical option 'allowdnsmismatch' in line %d", linecount);
		    continue;
		}
		if (*val && strcmp(opt, "useverifieddns") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_UseVerifiedDns = 1;
		    else
			readerdef.rd_UseVerifiedDns = 0;
		    continue;
		}
		if (*val && strcmp(opt, "denymismatcheddns") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_DenyMismatchedDns = 1;
		    else
			readerdef.rd_DenyMismatchedDns = 0;
		    continue;
		}
		if (*val && strcmp(opt, "denynodns") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_DenyNoDns = 1;
		    else
			readerdef.rd_DenyNoDns = 0;
		    continue;
		}
		if (*val && strcmp(opt, "grouplog") == 0) {
		    if (*val == 'y' || *val == 'Y' || *val == '1')
			readerdef.rd_GroupLog = 1;
		    else
			readerdef.rd_GroupLog = 0;
		    continue;
		}
		syslog(LOG_ERR, "dreader.access: Invalid readerdef option : %s in line %d\n", opt, linecount);
		continue;
	}
	/* Start of a new definition */
	if (strcmp(opt, "vserverdef") == 0) {
	    copt = OPT_VSERVER;
	    bzero(&vserver, sizeof(vserver));
	    strncpy(vserver.vs_Name, val, sizeof(vserver.vs_Name) - 1);
	    continue;
	}
	if (strcmp(opt, "groupdef") == 0) {
	    copt = OPT_GROUPS;
	    bzero(&vserver, sizeof(vserver));
	    groupdef.gr_Name = zallocStr(&DnsMemPool, val);
	    groupdef.gr_Count = 0;
	    groupdef.gr_Groups = NULL;
	    continue;
	}
	if (strcmp(opt, "authdef") == 0) {
	    copt = OPT_AUTH;
	    bzero(&vserver, sizeof(vserver));
	    strncpy(authdef.au_Name, val, sizeof(authdef.au_Name) - 1);
	    continue;
	}
	if (strcmp(opt, "readerdef") == 0) {
	    copt = OPT_READERGRP;
	    bzero(&vserver, sizeof(vserver));
	    strncpy(readerdef.rd_Name, val, sizeof(readerdef.rd_Name) - 1);
	    readerdef.rd_ControlPost = 1;
	    readerdef.rd_GroupLog = 1;
	    continue;
	}
	if (strcmp(opt, "access") == 0) {
	    char *p;
	    char *q;

	    bzero(&accessdef, sizeof(accessdef));
	    copt = OPT_NONE;
	    p = strtok(val, " \t");
	    p = strtok(NULL, " \t");
	    if (p == NULL) {
		syslog(LOG_ERR, "dreader.access: Invalid readerdef for access %sin line %d", val, linecount);
		continue;
	    }
	    while (*p == ' ' || *p == '\t')
		p++;
	    if (!*p || strlen(p) > sizeof(accessdef.ad_Reader)) {
		syslog(LOG_ERR, "dreader.access: Invalid readerdef for access %sin line %d", val, linecount);
		continue;
	    }
	    strcpy(accessdef.ad_Reader, p);
	    p = strtok(val, ",");
	    p = strtok(NULL, ",");
	    for (q = val; q != NULL; q = p, p = strtok(NULL, ",")) {
		char *pp = q;
		if ((q = strchr(pp, '@')) != NULL) {
		    *q++ = '\0';
		    if (!*pp || strlen(pp) > sizeof(accessdef.ad_Pattern)) {
			syslog(LOG_ERR, "dreader.access: Bad length for ident user %s in line %d", q, linecount);
			continue;
		    }
		    strncpy(accessdef.ad_IdentUser, pp,
			sizeof(accessdef.ad_IdentUser) - 1);
		} else {
		    q = pp;
		}
		if (!*q || strlen(q) > sizeof(accessdef.ad_Pattern)) {
		    syslog(LOG_ERR, "dreader.access: Bad length for access %s in line %d", q, linecount);
		    continue;
		}
		strcpy(accessdef.ad_Pattern, q);
		fwrite(&accessdef, sizeof(accessdef), 1, raf);
	    }
	    continue;
	}
	syslog(LOG_ERR, "dreader.access: Invalid access definition : %s in line %d\n", opt, linecount);
	Access_LastUpdate = 0;
    }
    fclose(fp);
    fclose(vsf);
    fclose(grf);
    fclose(auf);
    fclose(rdf);
    fclose(raf);
    if (Access_LastUpdate == 0) {
	syslog(LOG_ERR, "Errors in dreader.access file - not updating");
	return;
    }
    /*
     * We create a lock file that means no other processes should make
     * updates from the cache. The lock file time is also checked to
     * cater for possible race conditions.
     */

    if ((tfp = open(PatDbExpand(DRAccessLockPat), O_RDWR|O_CREAT, 0644)) <= 0) {
	syslog(LOG_CRIT, "Cannot create access lock %s",
		PatDbExpand(DRAccessLockPat));
	Access_LastUpdate = 0;
	return;
    }
    if (xflock(tfp, XLOCK_SH) < 0) {
	if (DebugOpt)
	    printf("Unable to obtain access lock (%s)\n", strerror(errno));
	Access_LastUpdate = 0;
	return;
    }

    rename(vsnewbuf, PatDbExpand(DRVserverCachePat));
    rename(grnewbuf, PatDbExpand(DRGroupCachePat));
    rename(aunewbuf, PatDbExpand(DRAuthCachePat));
    rename(rdnewbuf, PatDbExpand(DRReaderCachePat));
    rename(ranewbuf, PatDbExpand(DRAccessCachePat));

    write(tfp, &tfp, 1);
    xflock(tfp, XLOCK_UN);
    close(tfp);
}

/*
 * ReadAccessCache(): mmap the various access cache files into memory
 *   and set the pointers in the accessmap structure. This will be
 *   used to set the pointers for each thread.
 *
 * This has amazing performance benefits and saves memory because
 * we use the static mmap area for all threads
 */
int
ReadAccessCache() {
    int vsf;
    int grf;
    int auf;
    int rdf;
    int raf;
    int tfp;
    struct stat st;
    int i;
    time_t lastmod;
    int fail;

    if (DebugOpt > 1)
	printf("Checking access cache\n");

    if (stat(PatDbExpand(DRAccessLockPat), &st) != 0 ||
		st.st_mtime <= Cache_LastUpdate) {
	if (DebugOpt)
	    printf("No lock or not changed\n");
	return(0);
    }
    lastmod = st.st_mtime;

    /* Obtain a lock so that we know it won't disappear from under us */

    if ((tfp = open(PatDbExpand(DRAccessLockPat), O_RDONLY)) <= 0) {
	syslog(LOG_ERR, "Cannot open access lock %s",
		PatDbExpand(DRAccessLockPat));
	return(0);
    }
    if (hflock(tfp, 0, XLOCK_SH) < 0) {
	if (DebugOpt)
	    printf("No lock or not changed\n");
	return(0);
    }

    if (DebugOpt)
	printf("Loading access cache\n");

    fail = 0;
    /* We must be able to load all the files at	the same time */
    if ((vsf = open(PatDbExpand(DRVserverCachePat), O_RDONLY)) <= 0) {
	syslog(LOG_CRIT, "Unable to open %s", PatDbExpand(DRVserverCachePat));
	fail = 1;
    }
    if ((grf = open(PatDbExpand(DRGroupCachePat), O_RDONLY)) <= 0) {
	syslog(LOG_CRIT, "Unable to open %s", PatDbExpand(DRGroupCachePat));
	close(vsf);
	fail = 1;
    }
    if ((auf = open(PatDbExpand(DRAuthCachePat), O_RDONLY)) <= 0) {
	syslog(LOG_CRIT, "Unable to open %s", PatDbExpand(DRAuthCachePat));
	close(vsf);
	close(grf);
	fail = 1;
    }
    if ((raf = open(PatDbExpand(DRAccessCachePat), O_RDONLY)) <= 0) {
	syslog(LOG_CRIT, "Unable to open %s", PatDbExpand(DRAccessCachePat));
	close(vsf);
	close(grf);
	close(auf);
	fail = 1;
    }
    if ((rdf = open(PatDbExpand(DRReaderCachePat), O_RDONLY)) <= 0) {
	syslog(LOG_CRIT, "Unable to open %s", PatDbExpand(DRReaderCachePat));
	close(vsf);
	close(grf);
	close(auf);
	close(raf);
	fail = 1;
    }

    /* Now that the files are open, we can clear the lock */
    hflock(tfp, 0, XLOCK_UN);
    close(tfp);

    if (fail) {
	if (accessmap.AccessList == NULL)
	    return(-2);
	return(-1);
    }

    /*
     * Preserve the old accessmap for later munmap
     * We do this in case there is a race condition that removes the
     * map too early.  (????)
     */
    memcpy(&oldaccessmap, &accessmap, sizeof(accessmap));
    oldaccessmap.VServerList = accessmap.VServerList;
    oldaccessmap.GroupsList = accessmap.GroupsList;
    oldaccessmap.AuthList = accessmap.AuthList;
    oldaccessmap.ReaderList = accessmap.ReaderList;
    oldaccessmap.GroupMap = accessmap.GroupMap;
    oldaccessmap.AccessList = accessmap.AccessList;

    accessmap.VServerList = NULL;
    accessmap.VServerCount = 0;
    accessmap.GroupsList = NULL;
    accessmap.GroupCount = 0;
    accessmap.GroupMap = NULL;
    accessmap.AuthList = NULL;
    accessmap.AuthCount = 0;
    accessmap.ReaderList = NULL;
    accessmap.ReaderCount = 0;
    accessmap.AccessList = NULL;
    accessmap.AccessCount = 0;

    fstat(vsf, &st);
    accessmap.VServerList = xmap(NULL, st.st_size, PROT_READ, MAP_SHARED, vsf, 0);
    accessmap.VServerCount = st.st_size / sizeof(Vserver);
    close(vsf);

    fstat(grf, &st);
    if (st.st_size > 0) {
	char *mapptr;
	GroupDef *gp;
	GroupDef *pgp = NULL;
	GroupList *gl;
	GroupList *pgl = NULL;

	accessmap.GroupMap = xmap(NULL, st.st_size, PROT_READ, MAP_SHARED, grf, 0);
	mapptr = accessmap.GroupMap;
	while (*mapptr) {
	    gp = zalloc(&DnsMemPool, sizeof(GroupDef));
	    gp->gr_Groups = NULL;
	    gp->gr_Next = NULL;
	    if (accessmap.GroupsList == NULL)
		accessmap.GroupsList = gp;
	    if (pgp)
		pgp->gr_Next = gp;
	    gp->gr_Name = mapptr;
	    mapptr += strlen(gp->gr_Name) + 1;
	    gp->gr_Count = atoi(mapptr);
	    mapptr += strlen(mapptr) + 1;
	    pgl = NULL;
	    for (i=0; i < gp->gr_Count; i++) {
		gl = zalloc(&DnsMemPool, sizeof(GroupList));
		if (gp->gr_Groups == NULL)
		    gp->gr_Groups = gl;
		if (pgl)
		    pgl->next = gl;
		gl->group = zallocStr(&DnsMemPool, mapptr);
		mapptr += strlen(gl->group) + 1;
		gl->next = NULL;
		pgl = gl;
	    }
	    pgp = gp;
	    accessmap.GroupCount++;
	}
    }
    close(grf);

    fstat(auf, &st);
    accessmap.AuthList = xmap(NULL, st.st_size, PROT_READ, MAP_SHARED, auf, 0);
    accessmap.AuthCount = st.st_size / sizeof(AuthDef);
    close(auf);

    fstat(rdf, &st);
    accessmap.ReaderList = xmap(NULL, st.st_size, PROT_READ, MAP_SHARED, rdf, 0);
    accessmap.ReaderCount = st.st_size / sizeof(ReaderDef);
    close(rdf);

    fstat(raf, &st);
    accessmap.AccessList = xmap(NULL, st.st_size, PROT_READ, MAP_SHARED, raf, 0);
    accessmap.AccessCount = st.st_size / sizeof(AccessDef);
    close(raf);

    Cache_LastUpdate = lastmod;
    return(1);
}

/*
 * Remove the old mmap entries now that all the threads have been updated
 */
void
ClearOldAccessMap(void) {
    if (DebugOpt)
	printf("Clearing old accessmap\n");
    if (oldaccessmap.VServerList != NULL)
	xunmap(oldaccessmap.VServerList, oldaccessmap.VServerCount * sizeof(Vserver));
    if (oldaccessmap.GroupMap != NULL) {
	xunmap(oldaccessmap.GroupMap, oldaccessmap.GroupMapSize);
	zfree(&DnsMemPool, oldaccessmap.GroupsList, sizeof(GroupDef));
    }
    if (oldaccessmap.AuthList != NULL)
	xunmap(oldaccessmap.AuthList, oldaccessmap.AuthCount * sizeof(AuthDef));
    if (oldaccessmap.ReaderList != NULL)
	xunmap(oldaccessmap.ReaderList, oldaccessmap.ReaderCount * sizeof(ReaderDef));
    if (oldaccessmap.AccessList != NULL)
	xunmap(oldaccessmap.AccessList, oldaccessmap.AccessCount * sizeof(AccessDef));
    bzero(&oldaccessmap, sizeof(oldaccessmap));
}

/*
 * Update all the mmap'ed authentication pointers because something chanaged
 */
void
SetAuthDetails(DnsRes *dres, char *which)
{
    int i;
    ReaderDef *rdp;
    Vserver *vsp;
    GroupDef *gdp;
    AuthDef *adp;

    if (DebugOpt)
	printf("Setting Auth Details for %s\n", which);
    for (rdp = accessmap.ReaderList, i = 0; i < accessmap.ReaderCount;
								i++, rdp++)
	if (strcmp(which, rdp->rd_Name) == 0)
	    break;
    /* The first reader access line is the default */
    if (i >= accessmap.ReaderCount)
	rdp = accessmap.ReaderList;
    dres->dr_ReaderDef = rdp;
    
    for (vsp = accessmap.VServerList, i = 0; i < accessmap.VServerCount;
								i++, vsp++)
	/*
	 * If we are on a virtual interface, only check for the specified
	 * vserver.
	 */
	if (*dres->dr_VServer) {
	    if (strcmp(dres->dr_VServer, vsp->vs_Name) == 0)
		break;
	} else if (strcmp(rdp->rd_Vserver, vsp->vs_Name) == 0)
	    break;
    /* The first VServer is always the default one */
    if (i >= accessmap.VServerCount)
	vsp = accessmap.VServerList;
    dres->dr_VServerDef = vsp;
    
    for (gdp = accessmap.GroupsList, i = 0; i < accessmap.GroupCount;
						i++, gdp = gdp->gr_Next) {
	if (strcmp(rdp->rd_Groups, gdp->gr_Name) == 0)
	    break;
    }
    /* The first group list is always everything */
    if (i >= accessmap.GroupCount)
	gdp = accessmap.GroupsList;
    dres->dr_GroupDef = gdp;
    
    for (adp = accessmap.AuthList, i = 0; i < accessmap.AuthCount; i++, adp++)
	if (strcmp(rdp->rd_Auth, adp->au_Name) == 0)
	    break;
    /* The first auth list is no access */
    if (i >= accessmap.AuthCount)
	adp = accessmap.AuthList;
    dres->dr_AuthDef = adp;

    if (dres->dr_ReaderDef->rd_Read)
	dres->dr_Flags |= DF_READ;
    else
	dres->dr_Flags &= ~DF_READ;
    if (dres->dr_ReaderDef->rd_Post)
	dres->dr_Flags |= DF_POST;
    else
	dres->dr_Flags &= ~DF_POST;
    if (dres->dr_ReaderDef->rd_Feed)
	dres->dr_Flags |= DF_FEED;
    else
	dres->dr_Flags &= ~DF_FEED;
    if (dres->dr_ReaderDef->rd_Status)
	dres->dr_Flags |= DF_STATUS;
    else
	dres->dr_Flags &= ~DF_STATUS;
    if (dres->dr_ReaderDef->rd_GroupLog)
	dres->dr_Flags |= DF_GROUPLOG;
    else
	dres->dr_Flags &= ~DF_GROUPLOG;

}

/*
 * This gets called for each thread to update the pointers if the
 * mmapped info had been updated
 */
void
UpdateAuthDetails(ForkDesc *desc)
{
    Connection *conn = desc->d_Data;

    if (DebugOpt)
	printf("Updating auth details\n");
    if (conn && !conn->co_Auth.dr_StaticAuth)
	SetAuthDetails(&conn->co_Auth, conn->co_Auth.dr_ReaderName);
}

/*
 * PrintAuthDetails(): Used for debugging - print all auth info for a thread
 */
void
PrintAuthDetails(DnsRes *dres)
{
   printf("dr_ReaderName: %s\n", dres->dr_ReaderName);
   printf("dr_Flags: %x\n", dres->dr_Flags);
   printf("rd_Name: %s\n", dres->dr_ReaderDef->rd_Name);
    printf(" rd_Read: %d\n", dres->dr_ReaderDef->rd_Read);
    printf(" rd_Post: %d\n", dres->dr_ReaderDef->rd_Post);
    printf(" rd_Auth: %s\n", dres->dr_ReaderDef->rd_Auth);
    printf(" rd_Groups: %s\n", dres->dr_ReaderDef->rd_Groups);
    printf(" rd_Vserver: %s\n", dres->dr_ReaderDef->rd_Vserver);
    printf(" rd_MaxConnTotal: %d\n", dres->dr_ReaderDef->rd_MaxConnTotal);
    printf(" rd_MaxConnPerHost: %d\n", dres->dr_ReaderDef->rd_MaxConnPerHost);
    printf(" rd_MaxConnPerUser: %d\n", dres->dr_ReaderDef->rd_MaxConnPerUser);
    printf(" rd_MaxConnPerGroup: %d\n", dres->dr_ReaderDef->rd_MaxConnPerGroup);
    printf(" rd_MaxConnPerVs: %d\n", dres->dr_ReaderDef->rd_MaxConnPerVs);
    printf(" rd_RateLimit: %d\n", dres->dr_ReaderDef->rd_RateLimit);
    printf(" rd_RateLimitTax: %d\n", dres->dr_ReaderDef->rd_RateLimitTax);
    printf(" rd_IdleTimeout: %ld\n", dres->dr_ReaderDef->rd_IdleTimeout);
    printf(" rd_SessionTimeout: %ld\n", dres->dr_ReaderDef->rd_SessionTimeout);
    printf(" rd_GroupLog: %d\n", dres->dr_ReaderDef->rd_GroupLog);
    printf(" rd_UseVerifiedDns: %d\n", dres->dr_ReaderDef->rd_UseVerifiedDns);
    printf(" rd_DenyMismatchedDns: %d\n", dres->dr_ReaderDef->rd_DenyMismatchedDns);
    printf(" rd_DenyNoDns: %d\n", dres->dr_ReaderDef->rd_DenyNoDns);
   printf("vs_Name: %s\n", dres->dr_VServerDef->vs_Name);
    printf(" vs_HostName: %s\n", dres->dr_VServerDef->vs_HostName);
    printf(" vs_PostPath: %s\n", dres->dr_VServerDef->vs_PostPath);
    printf(" vs_NoXrefHostUpdate: %d\n", dres->dr_VServerDef->vs_NoXrefHostUpdate);
    printf(" vs_NewsAdm: %s\n", dres->dr_VServerDef->vs_NewsAdm);
    printf(" vs_Org: %s\n", dres->dr_VServerDef->vs_Org);
    printf(" vs_AbuseTo: %s\n", dres->dr_VServerDef->vs_AbuseTo);
    printf(" vs_CryptPw: %s\n", dres->dr_VServerDef->vs_CryptPw);
    printf(" vs_Interface: %ld\n", dres->dr_VServerDef->vs_Interface);
   printf("au_Name: %s\n", dres->dr_AuthDef->au_Name);
   printf(" au_File: %s\n", dres->dr_AuthDef->au_File);
   printf(" au_Cdb: %s\n", dres->dr_AuthDef->au_Cdb);
   printf(" au_External: %s\n", dres->dr_AuthDef->au_External);
   printf(" au_Radius: %s\n", dres->dr_AuthDef->au_Radius);
   printf(" au_User: %s\n", dres->dr_AuthDef->au_User);
   printf(" au_Pass: %s\n", dres->dr_AuthDef->au_Pass);
   printf(" au_Realm: %s\n", dres->dr_AuthDef->au_Realm);
   printf(" au_Ident: %d\n", dres->dr_AuthDef->au_Ident);
   printf("gr_Name: %s (%d)\n", dres->dr_GroupDef->gr_Name, dres->dr_GroupDef->gr_Count);
   {
	GroupList *gp;
	int i;
	for (gp = dres->dr_GroupDef->gr_Groups, i = 0;
		 i < dres->dr_GroupDef->gr_Count; i++, gp = gp->next)
	    printf(" group: %s\n", gp->group);
   }
}
