
/*
 * LIB/HOSTAUTH.C
 *
 * (c)Copyright 1997-1998, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 *
 */

#include "defs.h"

Prototype int LoadHostAccess(time_t t, int force, int rebuildtime);
Prototype char *Authenticate(int fd, struct sockaddr_in *asin, char *label);

#define	TESTTYPE_NONE	0
#define	TESTTYPE_ANY	1
#define	TESTTYPE_IP	2
#define	TESTTYPE_FQDN	4

typedef struct HostAccess {
    char		entry[256];
    int			testtype;
    struct sockaddr_in 	ipaddr;
    int			iplen;
    char		label[256];
} HostAccess;

time_t HATDSec = 0;
time_t HAMTime = 0;
time_t HACMTime = 0;

/*
 * LoadAccess() - [re]load diablo.hosts file
 *
 * This load is to create a cache file of the diablo.hosts file
 * that contains the IP addresses of all the hosts found. This
 * makes authentication a lot quicker and easier.
 *
 * If a time of '0' is passed, we don't fork and don't use syslog
 * 
 */

int
LoadHostAccess(time_t t, int force, int rebuildtime)
{
    int timeChanged = 0;
    struct stat st = { 0 };
    pid_t pid;

    /*
     * check for diablo.hosts file modified once a minute
     */

    if (!force  && (t / 10 <= HATDSec))
	return(0);

    timeChanged = 1;

    HATDSec = t / 10;

    /* Don't do this if diablo.hosts hasn't been modified  or we are
     * within the rebuild time limit
     */
    if (!force && HACMTime + rebuildtime > t &&
		stat(PatDbExpand(DHostsCachePat), &st) == 0 &&
		stat(PatLibExpand(DiabloHostsPat), &st) == 0 &&
		st.st_mtime == HAMTime)
	return(0);

    if (stat(PatLibExpand(DiabloHostsPat), &st) == 0)
	HAMTime = st.st_mtime;
     HACMTime = t;

    /*
     * Do the DNS lookups in the background
     * The main loop does a wait to cleanup the children
     */
    if (t != 0)
	pid = fork();
    else
	pid = 0;
    if (pid == 0) {
	FILE *fi;
	FILE *fo;
	char buf[512];
	HostAccess HA;
	struct hostent *he;
	int i;
	char *s;
	char *l;
	char *p;
        int lockfd;

	fi = fopen(PatLibExpand(DiabloHostsPat), "r");

	if (fi == NULL) {
	    if (t != 0)
		syslog(LOG_EMERG, "%s: %s", PatLibExpand(DiabloHostsPat),
							strerror(errno));
	    else
		fprintf(stderr, "%s: %s\n", PatLibExpand(DiabloHostsPat),
							strerror(errno));
	    exit(1);
	}

	if ((lockfd = open(PatDbExpand(DHostsLockPat), O_RDWR|O_CREAT, 0644)) < 0) {
	    if (t != 0)
		syslog(LOG_INFO, "%s : Unable to create lockfile (%s)",
				PatDbExpand(DHostsLockPat), strerror(errno));
	    else
		printf("%s : Unable to create lockfile (%s)\n",
				PatDbExpand(DHostsLockPat), strerror(errno));
	    exit(0);
	    
	}
	if (xflock(lockfd, XLOCK_EX|XLOCK_NB) == -1) {
	    if (t != 0)
		syslog(LOG_INFO, "%s : Rebuild already in progress",
						PatDbExpand(DHostsCachePat));
	    else
		printf("%s : Rebuild already in progress\n",
						PatDbExpand(DHostsCachePat));
	    close(lockfd);
	    exit(0);
	}
	sprintf(buf, "%d", getpid());
	write(lockfd, buf, strlen(buf));

	snprintf(buf, sizeof(buf), "%s.new", PatDbExpand(DHostsCachePat));

	fo = fopen(buf, "w");
	if (fo == NULL) {
	    if (t != 0)
		syslog(LOG_EMERG, "%s: %s", buf, strerror(errno));
	    else
		fprintf(stderr, "%s: %s\n", buf, strerror(errno));
	    xflock(lockfd, XLOCK_UN);
	    exit(1);
	}

	if (t != 0)
	    syslog(LOG_INFO, "Rebuilding host cache");
	else
	    printf("Rebuilding host cache\n");

	while (fgets(buf, sizeof(buf), fi) != NULL) {

	    p = buf;
	    if (*p == '\n' || *p == '#')
		continue;
	    HA.testtype = TESTTYPE_NONE;
	    if (strncmp(p, "IP:", 3) == 0) {
		HA.testtype |= TESTTYPE_IP;
		p += 3;
	    } else if (strncmp(p, "FQDN:", 5) == 0) {
		HA.testtype |= TESTTYPE_FQDN;
		p += 5;
	    }
	    s = strtok(p, ": \t\r\n");	/* hostname	*/
	    if (s == NULL || !*s)
		continue;
            l = strtok(NULL, ": \t\r\n");	/* label	*/
	    if (l == NULL || !*l)
		continue;

	    if (strlen(s) > 255 || strlen(l) > 255)
		continue;

	    /*
	     * If this entry is an IP, then only match against an IP
	     * otherwise only match against the hostname
	     */
	    if (HA.testtype == TESTTYPE_NONE) {
		if (IsIpAddr(s))
		    HA.testtype |= TESTTYPE_IP;
		else
		    HA.testtype |= TESTTYPE_FQDN;
	    }

	    strcpy(HA.entry, s);
	    strcpy(HA.label, l);
	    memset(&HA.ipaddr, 0, sizeof(HA.ipaddr));
	    HA.iplen = 0;

	    /*
	     * Don't lookup IP for wildcards/CIDR
	     */
	    if (strchr(s, '*') != NULL || strchr(s, '?') != NULL ||
			strchr(s, '/') != NULL) {
		fwrite(&HA, sizeof(HA), 1, fo);
		continue;
	    }

	    if ((he = gethostbyname(s)) != NULL)
		for (i = 0; he->h_addr_list && he->h_addr_list[i] != NULL; ++i) {
		    const struct in_addr *haddr = (const void *)he->h_addr_list[i];
		    memcpy(&HA.ipaddr.sin_addr, haddr, he->h_length);
		    HA.iplen = he->h_length;
		    HA.testtype |= TESTTYPE_IP;
		    fwrite(&HA, sizeof(HA), 1, fo);
		}
	}
	fclose(fi);
	fclose(fo);

	/*
	 * Now rename the newly created cache to its correct name
	 */
	snprintf(buf, sizeof(buf), "%s.new", PatDbExpand(DHostsCachePat));
	if ((stat(buf, &st) == 0 && st.st_size > 0) ||
		stat(PatDbExpand(DHostsCachePat), &st) != 0) {
	    rename(buf, PatDbExpand(DHostsCachePat));
	} else {
	    unlink(buf);
	}
	xflock(lockfd, XLOCK_UN);
	close(lockfd);
	if (t != 0)
	    exit(0);
    }
    return(pid);
}

/*
 * AUTHENTICATION()	- authenticate a new connection
 *
 *   We follow the following procedure:
 *
 *	- if access entry is IP, only match with IP
 *	- if access entry is FQDN, only match with FQDN
 *	- if IP has no hostname, only match IP
 *	- if IP Fwd/Rev mismatch, only match IP
 *	- if IP has valid hname, match with IP or hname
 *
 *	Returns the hostname or IP if hostname not found/invalid
 */

char *
Authenticate(int fd, struct sockaddr_in *asin, char *label)
{
    char *hname = NULL;		/* authenticated hostname or NULL */
    char *pname = NULL;		/* IP address (as string)	  */
    int isValid = 0;

    /*
     * check reverse lookup verses forward lookup, set either hname or uname
     * or neither, but never both.
     */

    {
	struct hostent *he;

	he = gethostbyaddr(
	    (char *)&asin->sin_addr,
	    sizeof(asin->sin_addr),
	    AF_INET
	);

	if (he != NULL) {
	    hname = zallocStr(&SysMemPool, he->h_name);
	    if (TestForwardLookup(hname, asin) < 0) {
		zfreeStr(&SysMemPool, &hname);
		hname = NULL;
	    }
	}
    }

    /*
     * set pname
     */
    pname = zallocStr(&SysMemPool, inet_ntoa(asin->sin_addr));

    /*
     * Check IP against cache file of diablo.hosts. 
     */
    {
	FILE *fi = fopen(PatDbExpand(DHostsCachePat), "r");

	if (fi) {
	    HostAccess HA;

	    while (fread(&HA, sizeof(HA), 1, fi) == 1) {
		if (HA.iplen && (HA.testtype & TESTTYPE_IP) && memcmp(&asin->sin_addr, &HA.ipaddr.sin_addr, HA.iplen) == 0) {
		    isValid = 1;
		    break;
		} else if ((strchr(HA.entry, '/') != NULL) && (HA.testtype & TESTTYPE_IP) && CidrMatch(HA.entry, pname)) {
		    isValid = 2;
		    break;
		} else if (pname && (HA.testtype & TESTTYPE_IP) && WildCaseCmp(HA.entry, pname) == 0) {
		    isValid = 3;
		    break;
		} else if (hname && (HA.testtype & TESTTYPE_FQDN) && strcasecmp(HA.entry, hname) == 0) {
		    isValid = 4;
		    break;
		} else if (hname && (HA.testtype & TESTTYPE_FQDN) && WildCaseCmp(HA.entry, hname) == 0) {
		    isValid = 5;
		    break;
		}
	    }
	    if (isValid) {
		char *method = "UNKNOWN";
		if (isValid == 1)
		    method = "Exact IP match";
		if (isValid == 2)
		    method = "CIDR match";
		if (isValid == 3)
		    method = "Wildcmp IP match";
		if (isValid == 4)
		    method = "Exact Hostname match";
		if (isValid == 5)
		    method = "Wildcmp hostname match";
		syslog(LOG_DEBUG, "Matched %s (%s) to %s (label:%s) using %s",
					hname ? hname : "NOTFOUND",
					pname, HA.entry, HA.label, method);
		if (HA.label[0])
		    strcpy(label, HA.label);
		else
		    label[0] = 0;
	    }
	    fclose(fi);
	} else {
	    syslog(LOG_ERR, "%s file not found", PatDbExpand(DHostsCachePat));
	}
    }

    /*
     * If we have a valid connection, but we were
     * unable to resolve the FQDN, we make the FQDN
     * the IP address.
     */
    if (isValid) {
	if (hname == NULL) {
	    hname = pname;
	    pname = NULL;
	}
    } else {
	if (hname)
	    zfreeStr(&SysMemPool, &hname);
    }
    zfreeStr(&SysMemPool, &pname);
    return(hname);
}

void
DumpHostCache(char *fname)
{
   FILE *fi = fopen(fname, "r");

    if (fi) {
	HostAccess HA;
        char tt[20];

	while (fread(&HA, sizeof(HA), 1, fi) == 1) {
	    strcpy(tt, "");
	    if (HA.testtype & TESTTYPE_NONE)
		strcat(tt, "NONE:");
	    if (HA.testtype & TESTTYPE_IP)
		strcat(tt, "IP:");
	    if (HA.testtype & TESTTYPE_FQDN)
		strcat(tt, "H:");
	    if (HA.testtype & TESTTYPE_ANY)
		strcat(tt, "ANY:");
	    tt[strlen(tt) - 1] = '\0';
	    printf("%s label=%s type=%s ip=%s\n", HA.entry, HA.label, tt,
				(HA.testtype & TESTTYPE_IP) ? inet_ntoa(HA.ipaddr.sin_addr) : "NONE");
	}
    } else {
	fprintf(stderr, "Unable to open %s\n", fname);
    }
}
